// vi: ai ts=2 noet sw=2 tw=80 nohls :

#include <iostream>
#include <string>
#include <stdexcept>
#include <sstream>
#include <cmath>
#include <cstdlib>
using namespace std;

const char* help=
"AICD-isocut\n"
"    (C) Felix Koehler 2006\n"
"Program to splice an AICD .Isoober.inc file\n"
"to a sphere or along a plane.\n"
"Usage:    AICD-isocut {options} < input.inc > output.inc\n"
"Options: \n"
"     -r <radius>: Radius to center (sphere only), default: 2.0\n"
"     -o <x> <y> <z>: cartesian coordinates of origin\n"
"     -n <x> <y> <z>: normal vector (plane only)\n"
"     -m <max>: maximum length of an arrow\n"
"Examples:\n"
"  AICD-isocut -r 1.0 < input.inc > output.inc\n"
"     deletes all elements not within a radius of 1.0 around origin.\nn"
"\n"
"  AICD-isocut -n 1.0 0.0 0.0 -o 1.0 0.0 0.0 < input.inc > output.inc\n"
"     deletes all elements in front of the plane given by\n"
"     the normal vector n=(1,0,0) and the point p=(1,0,0)\n"
;

double radius=2.0;
double arrow_maxlength=-1;
int statistics_arrows_deleted = 0;
double ox=0.0;
double oy=0.0;
double oz=0.0;
double nx=0.0;
double ny=0.0;
double nz=0.0;

typedef bool (*splice_func)(double, double, double);
bool splice_sphere (double px, double py, double pz)
{
//Calculate the distance from origin of point P. P is inside, if
//within a given radius.
	double ds= (px-ox)*(px-ox)
						+(py-oy)*(py-oy)
						+(pz-oz)*(pz-oz);
	if (ds <= radius*radius) return true;
	else return false;
}
bool splice_plane (double px, double py, double pz)
{
//Calculate the sign of the scalar s that will give the projection point
//P' of P on the plane given by the point O and the normalised
//normal vector n in the equation P' = P + s*n
// sign(s) = sign (n * PO)
	double sign_s = (nx*(ox-px) + ny*(oy-py) + nz*(oz-pz));
	if (sign_s>=0) return false;
	else return true;
}
splice_func is_inside=splice_sphere;


void parse(int argc, char**argv) {
	int i=0;
	while (++i<argc) {
		if (argv[i][0]=='-')
		{
			switch (argv[i][1])
			{
				case 'r':
					radius=atof(argv[++i]);
					break;
				case 'm':
					arrow_maxlength=atof(argv[++i]);
					break;
				case 'o':
					if (i+3 >= argc) throw std::runtime_error (
						"-o option: need three values to initialize coordinates of origin");
					ox=atof(argv[++i]);
					oy=atof(argv[++i]);
					oz=atof(argv[++i]);
					break;
				case 'n':
					if (i+3 >= argc) throw std::runtime_error (
						"-n option: need three values to initialize components of normal vector");
					nx=atof(argv[++i]);
					ny=atof(argv[++i]);
					nz=atof(argv[++i]);
					is_inside=splice_plane;
					break;
				default:
					cerr << help << endl;
					exit(0);
			}
		} else
		{
			cerr << help << endl;
			exit(0);
		}
	}
}

void delete_outer_surface() {
	string line;
	int count=0;
	bool isosurface_started = false;
	bool pfeile_started = false;
	while (true) {
		getline (cin, line);
		++count;
		if (!cin) break;

		// prescan line, if triangle or Pfeil
		bool is_triangle=false;
		bool is_pfeil=false;
		bool is_smooth=false;
		if (!isosurface_started)
			if (line.find("Isooberflaeche")!=string::npos) isosurface_started=true;
		if (!pfeile_started)
			if (line.find("Pfeile")!=string::npos) pfeile_started=true;
		if (isosurface_started && line.find("triangle") != string::npos)
			is_triangle=true;
		if (isosurface_started && line.find("smooth_triangle") != string::npos) 
		{
			is_triangle=true;
			is_smooth=true;
		}
		if (pfeile_started && line.find("cylinder") != string::npos) is_pfeil=true;
		if (!is_pfeil && !is_triangle) {
			cout << line.c_str() << "\n";
			continue;
		}
		bool ignore_line=true;

		// check if one point of triangle inside
		if (is_triangle)
		{
			std::istringstream iss(line);
			char ch;
			double x[3],y[3],z[3];
			for (int i=0; i<3; ++i) {
				while ((ch = iss.get())!='<')
					if (!iss) throw std::runtime_error (
						"Error reading Isooberflaeche entry");
				iss >> x[i] >> ch >> y[i] >> ch >> z[i];
				if (!iss) throw std::runtime_error (
						"Error reading Isooberflaeche entry");
				if (is_inside(x[i],y[i],z[i]))
				{
					ignore_line=false;
					break;
				}
				// smooth_triangles have extra vectors
				// PovRay doc:
				// SMOOTH_TRIANGLE:
				//	 smooth_triangle
				//			{
				//					<Corner_1>, <Normal_1>, <Corner_2>,
				//					<Normal_2>, <Corner_3>, <Normal_3>
				//					[OBJECT_MODIFIER...]
				//			}
				// this means we have to read and forget the normal vectors
				if (!is_smooth) continue;
				while ((ch = iss.get())!='<')
					if (!iss) throw std::runtime_error (
						"Error reading smooth Isooberflaeche entry");
				iss >> x[i] >> ch >> y[i] >> ch >> z[i];
				if (!iss) throw std::runtime_error (
						"Error reading smooth Isooberflaeche entry");
			}
		}

		// check, if starting point of Pfeil is within radius
		if (is_pfeil) {
			std::istringstream iss(line);
			char ch;
			double x,y,z;

			while ((ch = iss.get())!='<')
				if (!iss) throw std::runtime_error (
					"Error reading Pfeile entry");
			iss >> x >> ch >> y >> ch >> z;
			if (!iss) throw std::runtime_error (
					"Error reading Pfeile entry");
			if (is_inside(x,y,z)) ignore_line=false;

			// Check if arrow is not too long
			string s;
			double laenge=0.0;
			while (iss) {
				iss >> s;
				if (s=="Lnge:") {
					iss >> laenge;
					if (arrow_maxlength>0 && laenge>arrow_maxlength) {
						++statistics_arrows_deleted;
						ignore_line=true;
					}
					break;
				}
			}
		}

		// line has qualified to be copied. There's a bug in Daniels programs,
		// which produce '\0' chars in files. Delete these.
		//
		// Alos account for first line having a { that needs to be reproduced
		if (ignore_line)
		{
			if (line[0]=='{') cout << "{\n";
			continue;
		}
		for (int i=0; i<line.length(); ++i)
			if (line[i]==0) line[i]=' '; 
		cout << line << '\n';
	}
}

int main (int argc, char **argv) {
	ios::sync_with_stdio (false);
	try {
		cerr << "This is AICD-isocut...." << "\n";
		parse (argc, argv);
		cerr << "  arrow_maxlength = " << arrow_maxlength << "\n";
		cerr << "  Point of origin:  " << ox << " " << oy << " " << oz << "\n";
		if (is_inside == splice_sphere)
			cerr << "  Maximum distance: " << radius << "\n";
		else if (is_inside == splice_plane)
			cerr << "  Normal vector: (" << nx << ' ' << ny << ' ' << nz << ")\n";
		delete_outer_surface();
		cerr << "Number of deleted arrows: " << statistics_arrows_deleted << "\n";
	} catch (exception e) {
		cerr << e.what() << endl;
		return 1;
	}
	return 0;
}
